/*******************************************************}
{                                                       }
{       RichView                                        }
{       Demo: how to create MIME document.              }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************/


/*
  This demo shows how to create MIME (Multipurpose Internet Mail Extensions)
  document.
  These documents contain both HTML and images in one file.
  File extension:
  *.eml - e-mail file, can be opened, for example, by Outlook Express.
  *.mht - web archive, can be opened, for example, by Internet Explorer.
  These files are essentially the same.

  Additional information about MIME:
  RFC 2045
  http://www.ietf.org/rfc/rfc2045.txt
  RFC 2046
  http://www.ietf.org/rfc/rfc2046.txt
  RFC 2047
  http://www.ietf.org/rfc/rfc2047.txt
  RFC 2048
  http://www.ietf.org/rfc/rfc2048.txt
  RFC 2049
  http://www.ietf.org/rfc/rfc2049.txt
*/

/*
   This demo uses System.NetEncoding unit for base64 encoding.
   It is included in C++Builder XE7 and newer.
*/


//---------------------------------------------------------------------------

#include <vcl.h>
#include <stdio.h>
#pragma hdrstop

#include "Unit1.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "RichView"
#pragma link "RVEdit"
#pragma link "RVScroll"
#pragma link "RVStyle"
#pragma resource "*.dfm"
TForm1 *Form1;

TRVUnicodeString GetImageContentType(TGraphic* gr)
{
  switch(RVGraphicHandler->GetGraphicType(gr))
  {
    case rvgtBitmap:
      return "image/bmp";
	case rvgtIcon:
	  return "image/vnd.microsoft.icon";
	case rvgtMetafile:
	  return "application/octet-stream";
	case rvgtJPEG:
	  return "image/jpeg";
	case rvgtPNG:
	  return "image/png";
	case rvgtGIF:
	  return "image/gif";
    case rvgtTIFF:
      return "image/tiff";
    default:
      throw Exception("Unknown graphic format");
  }

}
//---------------------------------------------------------------------------
// THTMLImageItem
//---------------------------------------------------------------------------
__fastcall THTMLImageItem::THTMLImageItem(TCollection* Collection)
: TCollectionItem(Collection)
{
  Stream = new TMemoryStream;
}
//---------------------------------------------------------------------------
__fastcall THTMLImageItem::~THTMLImageItem()
{
  delete Stream;
}
//---------------------------------------------------------------------------
// THTMLImagesCollection
//---------------------------------------------------------------------------
__fastcall THTMLImagesCollection::THTMLImagesCollection()
: TCollection(__classid(THTMLImageItem))
{
}
//---------------------------------------------------------------------------
THTMLImageItem* __fastcall THTMLImagesCollection::GetItem(int Index)
{
  return (THTMLImageItem*)(TCollection::GetItem(Index));
}
//---------------------------------------------------------------------------
void __fastcall THTMLImagesCollection::SetItem(int Index, const THTMLImageItem* Value)
{
  TCollection::SetItem(Index, (TCollectionItem*)Value);
}
//---------------------------------------------------------------------------
// TForm1
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
        : TForm(Owner)
{

}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormCreate(TObject *Sender)
{
  HTMLImages = new THTMLImagesCollection;

  RichViewEdit1->Clear();
  RichViewEdit1->AddBreak(4, rvbsLine, clBlue);
  RichViewEdit1->AddNL("MIME File Test", 1, 1);
  Graphics::TBitmap *bmp = new Graphics::TBitmap;
  bmp->Assign(Image1->Picture->Graphic);
  RichViewEdit1->AddPicture("", bmp, 1);
  RichViewEdit1->AddNL("This example shows how to save HTML and images in one file.", 0, 0);
  RichViewEdit1->AddBreak(4, rvbsLine, clBlue);
  RichViewEdit1->Format();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormDestroy(TObject *Sender)
{
  delete HTMLImages;
}
//---------------------------------------------------------------------------
/* RichViewEdit1->OnSaveImage2: storing image in the HTMLImages collection instead
  of files. */
void __fastcall TForm1::RichViewEdit1SaveImage2(TCustomRichView *Sender,
      TGraphic *Graphic, TRVSaveFormat SaveFormat, const TRVUnicodeString Path,
      const TRVUnicodeString ImagePrefix, int &ImageSaveNo,
      TRVUnicodeString &Location, bool &DoDefault)
{
  if (SaveFormat!=rvsfHTML)
    return;

  TGraphic* gr;

  if (!Graphic->InheritsFrom(__classid(TJPEGImage)) &&
      !RVGraphicHandler->IsHTMLGraphic(Graphic))
  {
    TBitmap* bmp = RVGraphicHandler->ToBitmap(Graphic);
    gr = new TJPEGImage;
    gr->Assign(bmp);
    delete bmp;
  }
  else
    gr = Graphic;

  Location = RVFormatW("image%d.%s",
    ARRAYOFCONST((ImageSaveNo,  RVGraphicHandler->GetGraphicExt(gr))));
  ImageSaveNo++;
  THTMLImageItem* Item = (THTMLImageItem*)(HTMLImages->Add());
  gr->SaveToStream(Item->Stream);
  Item->Name = Location;
  Item->ContentType = GetImageContentType(gr);
  Location = "cid:"+Location;
  DoDefault = false;
  if (gr != Graphic)
	delete gr;
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Button1Click(TObject *Sender)
{

  if (! SaveDialog1->Execute())
	return;

  HTMLImages->Clear();

  // saving HTML
  TRVUnicodeString sw = "";
  TRVAnsiString s;

  TStringStream* StreamIn = new TStringStream(sw);
  RichViewEdit1->SaveHTMLToStreamEx(StreamIn, "", "Web Archive Demo", "", "", "", "",
	TRVSaveOptions() << rvsoUseCheckpointsNames << rvsoUTF8);
  TMemoryStream* StreamOut = new TMemoryStream;
  StreamIn->Position = 0;
  TNetEncoding::Base64->Encode(StreamIn, StreamOut);
  delete StreamIn;
  StreamOut->Position = 0;
  s.SetLength(StreamOut->Size);
  StreamOut->ReadBuffer(s.c_str(), s.Length());
  delete StreamOut;
  // now s contains HTML file without images, base64-encoded. saving it in MIME
  AnsiString boundary = "----=_BOUNDARY_LINE_";
  AnsiString FileName = SaveDialog1->FileName;
  FILE* F = fopen(FileName.c_str(), "wt");
  fprintf(F, "From: <Saved by TRICHVIEW.COM Demo>\n"
	"Subject: MIME Test\n"
	"MIME-Version: 1.0\n"
	"Content-Type: multipart/related;\n"
	"\tboundary=\"");
  fprintf(F, boundary.c_str());
  fprintf(F, "\"\nX-Unsent: 1\n\n"
	"This is a multi-part message in MIME format.\n\n"
	"--");
  fprintf(F, boundary.c_str());
  fprintf(F, "\nContent-Type: text/html;\n"
	"\tcharset=\"utf-8\"\n"
	"Content-Transfer-Encoding: base64\n\n");
  fprintf(F, s.c_str());
  fprintf(F, "\n");
  // saving images
  for (int i=0; i<HTMLImages->Count; i++)
  {
	fprintf(F, "\n--");
	fprintf(F, boundary.c_str());
	fprintf(F, "\n");
	fprintf(F, "Content-Type: ");
	s = UTF8Encode(HTMLImages->Items[i]->ContentType);
	fprintf(F, s.c_str());
	fprintf(F, "\n\tName=\"");
	s = UTF8Encode(HTMLImages->Items[i]->Name);
	fprintf(F, s.c_str());
	fprintf(F,  "\"\nContent-Transfer-Encoding: base64\n"
	  "Content-ID: <");
    fprintf(F, s.c_str());
	fprintf(F, ">\n\n");
	HTMLImages->Items[i]->Stream->Position = 0;
	StreamOut = new TMemoryStream;
	TNetEncoding::Base64->Encode(HTMLImages->Items[i]->Stream, StreamOut);
	StreamOut->Position = 0;
	s.SetLength(StreamOut->Size);
	StreamOut->ReadBuffer(s.c_str(), s.Length());
	delete StreamOut;
	fprintf(F, s.c_str());
	fprintf(F, "\n");
  }
  fprintf(F, "--");
  fprintf(F, boundary.c_str());
  fprintf(F, "--\n");
  fclose(F);
  HTMLImages->Clear();
}
//---------------------------------------------------------------------------