
{*******************************************************}
{                                                       }
{       RichView                                        }
{       Demo: how to create MIME document.              }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}


{
  Version 3.1
  Requires Indy 10+
  Requires TRichView 17.3+
  Requires a Unicode version of Delphi (2009 or newer).


  This demo shows how to send HTML e-mail.
  This e-mail includes text representation, HTML, images and attached files.

}

unit MainFrm;

interface

uses
  // Standard units
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Jpeg, ExtCtrls,
  // TRichView units
  RVStyle, RVScroll, RichView, RVEdit, RVGrHandler,
  RVGifAnimate2007, CRVData, RVUni, RVStrFuncs, RVFuncs, RVTypes,
  // Indy units
  IdMessage, IdBaseComponent, IdComponent,
  IdTCPConnection, IdTCPClient, IdMessageClient, IdSMTP,
  IdExplicitTLSClientServerBase, IdSMTPBase,
  IdText, IdAttachmentMemory, IdAttachmentFile, IdMessageCoderMIME;

type

  { This class will be used to store information about one picture while
    saving to HTML }
  THTMLImageItem = class (TCollectionItem)
    public
      Stream: TMemoryStream;
      Name, ContentType: TRVUnicodeString;
      constructor Create(Collection: TCollection); override;
      destructor Destroy; override;
  end;
  { Collection of THTMLImageItem }
  THTMLImagesCollection = class (TCollection)
  private
    function GetItem(Index: Integer): THTMLImageItem;
    procedure SetItem(Index: Integer; const Value: THTMLImageItem);
    public
      constructor Create;
      property Items[Index: Integer]: THTMLImageItem read GetItem write SetItem; default;
  end;

  TfrmMain = class(TForm)
    RVStyle1: TRVStyle;
    btnSend: TButton;
    Image1: TImage;
    IdSMTP1: TIdSMTP;
    IdMessage1: TIdMessage;
    gbFrom: TGroupBox;
    lblFromName: TLabel;
    txtFromName: TEdit;
    lblFromEMail: TLabel;
    txtFromEMail: TEdit;
    gbTo: TGroupBox;
    lblToName: TLabel;
    lblToEMail: TLabel;
    txtToName: TEdit;
    txtToEMail: TEdit;
    gbMessage: TGroupBox;
    Label1: TLabel;
    txtSubject: TEdit;
    RichViewEdit1: TRichViewEdit;
    Label2: TLabel;
    gbServer: TGroupBox;
    lblHost: TLabel;
    lblPort: TLabel;
    txtHost: TEdit;
    txtPort: TEdit;
    lblUserID: TLabel;
    lblPassword: TLabel;
    txtPassword: TEdit;
    txtUserID: TEdit;
    gbFiles: TGroupBox;
    txtFiles: TMemo;
    Label3: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RichViewEdit1SaveImage2(Sender: TCustomRichView;
      Graphic: TGraphic; SaveFormat: TRVSaveFormat; const Path,
      ImagePrefix: TRVUnicodeString; var ImageSaveNo: Integer;
      var Location: TRVUnicodeString;
      var DoDefault: Boolean);
    procedure btnSendClick(Sender: TObject);
  private
    { Private declarations }
    HTMLImages: THTMLImagesCollection;
    Attachments: TStringList;
    function PrepareAttachments: Boolean;
    procedure BuildEmail;
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

const
  crlf = #13#10;

implementation

{$R *.dfm}

function GetImageContentType(gr: TGraphic): TRVUnicodeString;
begin
  case RVGraphicHandler.GetGraphicType(gr) of
    rvgtBitmap:
      Result := 'image/bmp';
    rvgtIcon:
      Result := 'image/vnd.microsoft.icon';
    rvgtMetafile:
      Result := 'application/octet-stream';
    rvgtJPEG:
      Result := 'image/jpeg';
    rvgtPNG:
      Result := 'image/png';
    rvgtGIF:
      Result := 'image/gif';
    rvgtTIFF:
      Result := 'image/tiff';
    else
      raise Exception.Create('Unknown graphic format');
  end;
end;

{=============================== THTMLImageItem ===============================}
constructor THTMLImageItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  Stream := TMemoryStream.Create;
end;
{------------------------------------------------------------------------------}
destructor THTMLImageItem.Destroy;
begin
  Stream.Free;
  inherited;
end;
{============================ THTMLImagesCollection ===========================}
constructor THTMLImagesCollection.Create;
begin
  inherited Create(THTMLImageItem);
end;
{------------------------------------------------------------------------------}
function THTMLImagesCollection.GetItem(Index: Integer): THTMLImageItem;
begin
  Result := THTMLImageItem(inherited GetItem(Index));
end;
{------------------------------------------------------------------------------}
procedure THTMLImagesCollection.SetItem(Index: Integer;
  const Value: THTMLImageItem);
begin
  inherited SetItem(Index, Value);
end;
{================================= TForm1 =====================================}
procedure TfrmMain.FormCreate(Sender: TObject);
var bmp: TBitmap;
begin
  // This is a list of file names to attach to the message
  Attachments := TStringList.Create;
  // Creating a sample document
  RichViewEdit1.Clear;
  RichViewEdit1.AddBreak(4, rvbsLine, clBlue);
  RichViewEdit1.AddNL('HTML E-mail Test', 1, 1);
  bmp := TBitmap.Create;
  bmp.Assign(Image1.Picture.Graphic);
  RichViewEdit1.AddPicture('', bmp, 1, rvvaBaseline);
  RichViewEdit1.AddNL('This example shows how to send HTML e-mail using Indy 10 components.', 0, 0);
  RichViewEdit1.AddBreak(4, rvbsLine, clBlue);
  RichViewEdit1.Format;
end;
{------------------------------------------------------------------------------}
procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  Attachments.Free;
end;
{------------------------------------------------------------------------------}
// Copying from txtFiles to Attachments. Checking if all files exist
function TfrmMain.PrepareAttachments: Boolean;
var
  i: Integer;
  FileName: String;
begin
  Result := False;
  Attachments.Clear;
  for i := 0 to txtFiles.Lines.Count-1 do
  begin
    FileName := Trim(txtFiles.Lines[i]);
    if FileName='' then
      continue;
    if not FileExists(FileName) then
    begin
      ShowMessage('"'+FileName+'" does not exist!');
      exit;
    end;
    Attachments.Add(FileName);
  end;
  Result := True;
end;

{------------------------------------------------------------------------------}
{ RichViewEdit1.OnSaveImage2: storing image in the HTMLImages collection instead
  of files. }
procedure TfrmMain.RichViewEdit1SaveImage2(Sender: TCustomRichView;
  Graphic: TGraphic; SaveFormat: TRVSaveFormat; const Path,
  ImagePrefix: TRVUnicodeString; var ImageSaveNo: Integer;
  var Location: TRVUnicodeString; var DoDefault: Boolean);
var gr: TGraphic;
    bmp: TBitmap;
begin
  if SaveFormat<>rvsfHTML then
    exit;
  if not (Graphic is TJPEGImage) and not RVGraphicHandler.IsHTMLGraphic(Graphic) then
  begin
    bmp := RVGraphicHandler.ToBitmap(Graphic);
    try
      gr := TJPEGImage.Create;
      gr.Assign(bmp);
    finally
      bmp.Free;
    end;
  end
  else
    gr := Graphic;

  Location := Format('image%d.%s', [ImageSaveNo, GraphicExtension(TGraphicClass(gr.ClassType))]);
  inc(ImageSaveNo);
  with HTMLImages.Add as THTMLImageItem do
  begin
    gr.SaveToStream(Stream);
    Name := Location;
    ContentType := GetImageContentType(gr);
  end;
  Location := 'cid:'+Location;
  DoDefault := False;
  if gr<>Graphic then
    gr.Free;
end;
{------------------------------------------------------------------------------}
procedure TfrmMain.BuildEmail;
var Stream: TMemoryStream;
    ws: String;
    s: TRVRawByteString;
    i: Integer;
    txtpart, htmpart, txthtmpart, txthtmlimgpart: TIdText;
    imgpart: TIdAttachmentMemory;
begin
  // saving text
  txtpart := TIdText.Create(IdMessage1.MessageParts);
  Stream := TMemoryStream.Create;
  RichViewEdit1.SaveTextToStreamW('', Stream, 80, False, False);
  Stream.Position := 0;
  SetLength(ws, Stream.Size div 2);
  if Stream.Size<>0 then
    Stream.ReadBuffer(Pointer(ws)^, Stream.Size);
  Stream.Free;
  txtpart.Body.Text := ws;
  txtpart.ContentType := 'text/plain';
  txtpart.ContentTransfer := 'quoted-printable';
  txtpart.CharSet := 'utf-8';
  // saving HTML
  htmpart := TIdText.Create(IdMessage1.MessageParts);
  Stream := TMemoryStream.Create;
  // Note: rvsoInlineCSS generates larger HTML file, but it looks better in
  // online e-mail clients
  RichViewEdit1.SaveHTMLToStreamEx(Stream, '', txtSubject.Text, '', '', '', '',
    [rvsoUseCheckpointsNames, rvsoUTF8, rvsoInlineCSS]);
  Stream.Position := 0;
  SetLength(s, Stream.Size);
  if Stream.Size<>0 then
    Stream.ReadBuffer(Pointer(s)^, Stream.Size);
  Stream.Free;
  htmpart.Body.Text := UTF8ToUnicodeString(s);
  htmpart.ContentType := 'text/html';
  htmpart.ContentTransfer := 'quoted-printable';
  htmpart.CharSet := 'utf-8';
  // combining text and HTML
  IdMessage1.ContentType := 'multipart/alternative';
  if (Attachments.Count=0) and (HTMLImages.Count=0) then
    exit;
  txthtmpart := TIdText.Create(IdMessage1.MessageParts);
  txthtmpart.ContentType := 'multipart/alternative';
  txthtmpart.Index := 0;
  txtpart.ParentPart := txthtmpart.Index;
  htmpart.ParentPart := txthtmpart.Index;
  // images
  for i := 0 to HTMLImages.Count-1 do begin
    HTMLImages[i].Stream.Position := 0;
    imgpart := TIdAttachmentMemory.Create(IdMessage1.MessageParts, HTMLImages[i].Stream);
    imgpart.ContentType := HTMLImages[i].ContentType;
    imgpart.ContentID := '<'+HTMLImages[i].Name+'>';
    imgpart.ContentDisposition := 'inline';
  end;
  // combining images and text+html
  IdMessage1.ContentType := 'multipart/related; type="multipart/alternative"';
  if Attachments.Count=0 then
    exit;
  if HTMLImages.Count>0 then begin
    txthtmlimgpart := TIdText.Create(IdMessage1.MessageParts);
    txthtmlimgpart.ContentType := 'multipart/related; type="multipart/alternative"';
    txthtmlimgpart.Index := 0;

    txthtmpart.ParentPart := txthtmlimgpart.Index;
    for i := IdMessage1.MessageParts.Count-1-HTMLImages.Count to IdMessage1.MessageParts.Count-1 do
      IdMessage1.MessageParts[i].ParentPart := txthtmlimgpart.Index;

    txtpart.ParentPart := txthtmpart.Index;
    htmpart.ParentPart := txthtmpart.Index;
  end;
  // files
  IdMessage1.ContentType := 'multipart/mixed';
  for i := 0 to Attachments.Count-1 do
    TIdAttachmentFile.Create(IdMessage1.MessageParts, Attachments[i]);
end;
{------------------------------------------------------------------------------}
{ Sending using Indy components }
procedure TfrmMain.btnSendClick(Sender: TObject);
begin
  if not PrepareAttachments then
    exit;
  btnSend.Enabled := False;
  btnSend.Caption := 'Wait...';
  try

    IdMessage1.Clear;
    IdMessage1.From.Address := txtFromEMail.Text;
    IdMessage1.From.Name := txtFromName.Text;

    with IdMessage1.Recipients.Add do begin
      Address := txtToEMail.Text;
      Name := txtToName.Text;
    end;

    IdMessage1.Subject := txtSubject.Text;

    with IdMessage1.ReplyTo.Add do begin
      Address := txtFromEMail.Text;
      Name := txtFromName.Text;
    end;

    HTMLImages := THTMLImagesCollection.Create;
    try
      BuildEmail;
    finally
      HTMLImages.Free;
    end;

    IdSMTP1.MailAgent := 'Demo HTML Mailer';
    IdSMTP1.Host := txtHost.Text;
    IdSMTP1.Port := StrToIntDef(txtPort.Text, 25);
    IdSMTP1.Username := txtUserID.Text;
    IdSMTP1.Password := txtPassword.Text;

    IdSMTP1.Connect;
    try
      IdSMTP1.Send(IdMessage1);
    finally
      IdSMTP1.Disconnect;
    end;

    ShowMessage('Success!');
  finally
    btnSend.Enabled := True;
    btnSend.Caption := 'Send';
  end;
end;

end.